/*
Copyright (c) 2011, Tom Van Cutsem, Vrije Universiteit Brussel
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the Vrije Universiteit Brussel nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.*/
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.RecursiveAction;
import java.util.concurrent.RecursiveTask;

/**
 * Divide-and-conquer parallel map over an array of ints.
 * 
 * Based on Dan Grossman's lecture notes
 * "A Sophomoric Introduction to Shared-Memory Parallelism and Concurrency"
 * {@link http://www.cs.washington.edu/homes/djg/teachingMaterials/}
 * 
 * @author tvcutsem
 */
public class FJMap extends RecursiveAction {
	
	/**
	 * Users of {@link FJMap} should use a {@link Fun} instance to map
	 * source to target elements.
	 */
	public static interface Fun {
		public int map(int index, int source);
	}
	
	private static final int SEQUENTIAL_CUTOFF = 1000;
	
	final int lo;
	final int hi;
	final int[] in;
	final int[] out;
	final Fun fun;
	
	FJMap(int[] in, int[] out, int lo, int hi, Fun fn) {
		this.in  = in;
		this.lo  = lo;
		this.hi  = hi;
		this.out = out;
		this.fun = fn;
	}
	
	protected void compute(){// return answer
    if ((hi - lo) < SEQUENTIAL_CUTOFF) {
    	for (int i = lo; i < hi; i++) {
    		out[i] = fun.map(i, in[i]);
    	}
    } else {
    	FJMap left = new FJMap(in,out,lo,(hi+lo)/2,fun);
    	FJMap right= new FJMap(in,out,(hi+lo)/2,hi,fun);
    	left.fork();
    	right.compute();
    	left.join();
    }
	}
	
	final static ForkJoinPool fjPool = new ForkJoinPool(); 
	
	static int[] pmap(int[] input, Fun fun) {
	  int[] output = new int[input.length];
	  fjPool.invoke(new FJMap(input,output,0,input.length,fun));
	  return output;
	}
	
	public static void main(String[] args) throws Exception {
		int SIZE = 10000 * 4;
		int[] input = new int[SIZE];
		for (int i = 0; i < input.length; i++) {
			input[i] = i;
		}
		long start = System.currentTimeMillis();
		Fun squarer = new Fun() {
			public int map(int idx, int i) {
				return i*i;
			}
		};
		int[] output = pmap(input, squarer);
		long end = System.currentTimeMillis();
		
		// validation
		for (int i = 0; i < output.length; i++) {
			int res = squarer.map(i, input[i]);
			if (output[i] != res)
				throw new RuntimeException("Wrong output for index " + i + " expected "+res + " got "+output[i]);
		}
		
		System.out.println("Map computed in "+(end-start)+" ms");
	}
	
}