/*
Copyright (c) 2011, Tom Van Cutsem, Vrije Universiteit Brussel
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the Vrije Universiteit Brussel nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.*/
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.RecursiveAction;

/**
 * Divide-and-conquer parallel pack over an array of ints.
 * 
 * Based on Dan Grossman's lecture notes
 * "A Sophomoric Introduction to Shared-Memory Parallelism and Concurrency"
 * {@link http://www.cs.washington.edu/homes/djg/teachingMaterials/}
 * 
 * @author tvcutsem
 */
public class FJPack extends RecursiveAction {
	
	private static final int SEQUENTIAL_CUTOFF = 1000;
	
	final int[] in;
	final int[] out;
	final int lo;
	final int hi;
	final int[] bitvec;
	final long[] bitsum;
	
	FJPack(int[] input, int[] output, int lo, int hi, int[] bitvec, long[] bitsum) {
		this.in     = input;
		this.out    = output;
		this.lo     = lo;
		this.hi     = hi;
		this.bitvec = bitvec;
		this.bitsum = bitsum;
	}
	
	protected void compute() {
		if ((hi - lo) < SEQUENTIAL_CUTOFF) {
			for (int i = lo; i < hi; i++) {
				if (bitvec[i] == 1) {
					out[(int)bitsum[i]-1] = in[i];
				}
			}
		} else {
			FJPack left = new FJPack(in,out,lo,(hi+lo)/2,bitvec,bitsum);
			FJPack right= new FJPack(in,out,(hi+lo)/2,hi,bitvec,bitsum);
			left.fork();
			right.compute();
			left.join();
		}
	}
	
	final static ForkJoinPool fjPool = new ForkJoinPool(); 
	
	public static interface Filter {
		public boolean filter(int i);
	}
	
	static int[] parallelPack(int[] input, final Filter filter) {
		final int[] bitvector = new int[input.length];
		// step 1: perform parallel map to produce bit vector of {0,1} indicating
		// whether the corresponding input element satisfies the filter
		fjPool.invoke(new FJMap(input, bitvector, 0, input.length, new FJMap.Fun() {
			public int map(int idx, int source) {
				return (filter.filter(source)) ? 1 : 0;
			}
		}));
		// step 2: perform parallel prefix sum on the bit vector
		final long[] bitsum = FJPrefixSum.parallelPrefixSum(bitvector);
		// step 3: perform parallel map to assign input elements to bitsum[i]-1 in the
		// output array if bitvector[i] is set
		int out_length = (int) bitsum[bitsum.length-1];
		final int[] output = new int[out_length];
		fjPool.invoke(new FJPack(input, output, 0, input.length, bitvector, bitsum));
		return output;
	}
	
	static int[] sequentialPack(int[] input, Filter filter) {
		int count = 0;
		for (int i = 0; i < input.length; i++) {
			if (filter.filter(input[i])) {
				count++;
			}
		}
		int[] output = new int[count];
		int idx = 0;
		for (int i = 0; i < input.length; i++) {
			if (filter.filter(input[i])) {
				output[idx++] = input[i];
			}
		}
		// assert(index==count);
		return output;
	}
	
	public static void main(String[] args) throws Exception {
		int SIZE = 10000 * 4;
		int[] input = new int[SIZE];
		for (int i = 0; i < input.length; i++) {
			input[i] = i;
		}
		long start = System.currentTimeMillis();
		Filter divisibleBy5 = new Filter() {
		    public boolean filter(int i) {
		    	try {
					Thread.sleep(10);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
				return (i % 5) == 0;
			}
		};
		int[] output = parallelPack(input, divisibleBy5);
		long end = System.currentTimeMillis();
		
		long tPar = (end-start);
		
		// validation
		start = System.currentTimeMillis();
		int[] seqOut = sequentialPack(input, divisibleBy5);
		end = System.currentTimeMillis();
		long tSeq = (end-start);
		
		for (int i = 0; i < output.length; i++) {
			if (output[i] != seqOut[i])
				throw new RuntimeException(
						"Wrong output for index " + i + " expected "+seqOut[i] + " got "+output[i]);
		}
		
		System.out.println("Pack computed in "+tPar+" ms, speedup = " + 1.0 * tSeq / tPar);
	}
	
}