/*
Copyright (c) 2011, Tom Van Cutsem, Vrije Universiteit Brussel
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the Vrije Universiteit Brussel nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.*/
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.RecursiveAction;
import java.util.concurrent.RecursiveTask;

/**
 * Parallel prefix sum using the Java Fork/Join framework.
 * 
 * Based on Dan Grossman's lecture notes
 * "A Sophomoric Introduction to Shared-Memory Parallelism and Concurrency"
 * {@link http://www.cs.washington.edu/homes/djg/teachingMaterials/}
 * 
 * @author tvcutsem
 */
public class FJPrefixSum extends RecursiveAction {
	
	private static final int SEQUENTIAL_CUTOFF = 1000;
	
	/**
	 * The {@link FJBuildNode} task will build a binary tree
	 * consisting of these Nodes. Child nodes have their left
	 * and right field set to <tt>null</tt>.
	 * 
	 * Each node stores the sum of all elements in range [lo,hi[.
	 */
	static class Node {
		final Node left;
		final Node right;
		final long sum;
		final int lo;
		final int hi;
		Node(Node l, Node r, long s, int low, int high) {
			left  = l;
			right = r;
			sum   = s;
			lo    = low;
			hi    = high;
		}
		boolean isLeaf() { return left == null; }
	}
	
	/**
	 * This task computes the first or "up" phase of a parallel prefix,
	 * computing a binary tree in parallel from the leaf nodes and returning
	 * the root. Each node contains the sum of all elements in its range.
	 */
	static class FJBuildNode extends RecursiveTask<Node> {

		final int[] in;
		final int lo;
		final int hi;
		
		FJBuildNode(int[] a, int l, int h) {
			in = a;
			lo = l;
			hi = h;
		}
		
		protected Node compute(){
	    if ((hi - lo) < SEQUENTIAL_CUTOFF) {
	    	// sequential sum from lo to hi
	    	long sum = 0;
			  for (int i = lo; i < hi; i++) {
				  sum += in[i];
			  }
	    	return new Node(null, null, sum, lo, hi);
	    } else {
	    	FJBuildNode left = new FJBuildNode(in,lo,(hi+lo)/2);
	    	FJBuildNode right= new FJBuildNode(in,(hi+lo)/2,hi);
	    	left.fork();
	    	Node rightNode = right.compute();
	    	Node leftNode = left.join();
	    	return new Node(leftNode, rightNode, leftNode.sum + rightNode.sum, lo, hi);
	    }
		}
		
	}

	
	final Node node;
  final long leftSum;
  final int[] in;
	final long[] out;
	
	FJPrefixSum(Node n, long ls, int[] input, long[] output) {
		node    = n;
		leftSum = ls;
		in      = input;
		out     = output;
	}
	
	protected void compute(){// return answer
	  if (node.isLeaf()) {
		  // sequential prefix sum from [lo,hi[
		  final int lo = node.lo;
		  final int hi = node.hi;
		  out[lo] = in[lo] + leftSum;
		  for (int i = lo + 1; i < hi; i++) {
		    out[i] = out[i-1] + in[i];
		  }
		  // if sequential cutoff = 1, would default to:
    	// out[node.lo] = in[node.lo] + leftSum;
    } else {
    	FJPrefixSum left = new FJPrefixSum(node.left,  leftSum, in, out);
    	FJPrefixSum right= new FJPrefixSum(node.right, leftSum + node.left.sum, in, out);
    	left.fork();
    	right.compute();
    	left.join();
    }
	}
	
	final static ForkJoinPool fjPool = new ForkJoinPool(); 
	
	static long[] parallelPrefixSum(int[] input) {
	  long[] output = new long[input.length];
	  Node root = fjPool.invoke(new FJBuildNode(input,0,input.length)); // step 1: up
	  fjPool.invoke(new FJPrefixSum(root, 0, input, output)); // step 2: down
	  return output;
	}
	
	static long[] sequentialPrefixSum(int[] input) {
		long[] output = new long[input.length];
		output[0] = input[0];
		for (int i = 1; i < output.length; i++) {
			output[i] = output[i-1] + input[i];
		}
		return output;
	}
	
	public static void main(String[] args) throws Exception {
		int SIZE = 100000;
		int[] array = new int[SIZE];
		for (int i = 0; i < array.length; i++) {
			array[i] = i;
		}
		long start = System.currentTimeMillis();
		long[] theSum = parallelPrefixSum(array);
		long end = System.currentTimeMillis();
		
		// test validity
		long[] theSum2 = sequentialPrefixSum(array);
		for (int i = 0; i < theSum2.length; i++) {
			if (theSum[i] != theSum2[i])
			  throw new RuntimeException(
					  "Wrong answer at index "+i+" expected: "+theSum2[i] + " got "+theSum[i]);
		}
		
		System.out.println("Prefix sum computed in "+(end-start)+" ms");
	}
	
}